==> faws-0.1.0.tar.gz: faws-0.1.0/faws/core.py <==
# -*- coding: utf-8 -*-
import itertools

import boto3

RESOURCES = ('ec2', 'elb', 'sg', 'roles', 'elasticache', 'asg')

ec2 = boto3.client('ec2')
elbc = boto3.client('elb')
iam = boto3.client('iam')
ec = boto3.client('elasticache')
asg = boto3.client('autoscaling')

all_auto_scaling_groups = []
all_ec_clusters = []
all_load_balancers = []
all_security_groups = []
all_ec2_instances = []
all_roles = []


def get_all_ec2_instances():
    return ec2.describe_instances()['Reservations']


def get_all_security_groups():
    return ec2.describe_security_groups()['SecurityGroups']


def get_all_load_balancers():
    return elbc.describe_load_balancers()['LoadBalancerDescriptions']


def request_all(client, funcname, topkey, MaxRecords=100):
    '''
    Helper function get all result, handle pagination, return list of things
    '''
    paginator = client.get_paginator(funcname)
    all_entities = []

    page_iterator = paginator.paginate(MaxRecords=MaxRecords)
    for page in page_iterator:
        all_entities.extend(page[topkey])

    return all_entities


def get_all_auto_scaling_groups():
    return request_all(asg, 'describe_auto_scaling_groups', 'AutoScalingGroups')


def get_all_cache_clusters():
    return request_all(ec, 'describe_cache_clusters', 'CacheClusters')


def get_all_roles():
    # TODO paginate
    roles = iam.list_roles(MaxItems=1000)
    assert len(roles['Roles']) < 1000, 'THere are more than 1000 roles, do paginate, lazy!'
    return roles['Roles']


def get_instances():
    '''
    Get all instances, save to global var to not query AWS anymore
    '''
    pass


def get_instance(instance_id):
    '''
    Get from local instead of call AWS
    '''
    pass


def instance_ids_to_ips(instance_ids):
    # type: List -> Dict[id->ip]
    return


def get_elbs(namefilter=None):
    '''
    Get ELBs and save to global
    '''
    # Use fitler to filter base on elb name
    pass


# TODO design func to get ELB health - related instances
# HOW TO quickly get list of instances that out of service? and/or because of Unhealthy ELB check?


def guess_service_name(lb_name):
    # NOTE: order, duplication matter
    for part in ('ec2-', 'elb-', 'prd-', 'elb-', 'prod-'):
        if lb_name.startswith(part):
            lb_name = lb_name[len(part):]

    # TODO: should -int is separated thing?
    if lb_name.endswith('-int'):
        lb_name = lb_name[:-4]

    # TODO: handle suffix number, like service-1 app-2... workaround when there are too much inbound rules
    if lb_name.endswith('-1'):
        lb_name = lb_name[:-2]

    return lb_name


def extract_elb_name_and_sgs_from_elb(lb):
    if isinstance(lb, str):
        # got pretend sg name, and get inbound of the SG
        sgs = [sg for sg in all_security_groups if sg['GroupName'].lower() == lb]
        inbound_sgs = []
        for sg in sgs:
            for inbound_rule in sg['IpPermissions']:
                inbound_sg_id = inbound_rule['UserIdGroupPairs'][0]['GroupId']
                inbound_sgs.append(inbound_sg_id)

        return lb, inbound_sgs
    else:
        return lb['LoadBalancerName'], lb['SecurityGroups']


def sgid_to_inbounds(sgid):
    # inbound rules -> security group ids -> security group name
    security_group = [sg for sg in all_security_groups if sg['GroupId'] == sgid]
    assert len(security_group) == 1, 'Should only 1 SG has given ID {}'.format(sgid)

    inbound_rules = security_group[0]['IpPermissions']

    internet_access = False
    groupsids = []
    for rule in inbound_rules:
        if rule['UserIdGroupPairs']:
            groupsids.append(rule['UserIdGroupPairs'])
        else:
            # UserIdGroupPairs empty means it is a CIDR rule - maybe check for internet incoming?
            if rule['FromPort'] in (80, 443):
                # TODO maybe check more
                internet_access = True

    sgs = [g['GroupId'] for L in groupsids for g in L]
    ret = {}
    try:
        if sgs:
            ret = {sg['GroupName']: sg['GroupId']
                   for sg in all_security_groups if sg['GroupId'] in sgs}
    except Exception as e:
        print(e)

    if internet_access:
        ret.update({'INTERNET': '1'})
    return ret


def find_security_groups_contain(service_name):
    service_name = service_name.lower()
    return [sg for sg in all_security_groups if service_name in sg['GroupName'].lower()]


def is_prod_sg(sg):
    return 'prd-' in sg['GroupName'] or 'prod-' in sg['GroupName']


def get_sg_name(sg):
    return sg['GroupName']


def get_main_security_groups(service_name):
    groups = [g for g in find_security_groups_contain(service_name) if is_prod_sg(g)]

    for sg in groups:
        inbounds = sgid_to_inbounds(sg['GroupId'])
        elb = {name: inbounds[name] for name in inbounds if 'elb-' in name}
        if elb:
            elb.update(
                {sg['GroupName']: sg['GroupId']}
            )
            return elb
    return {}


def elb_obj_to_all_inbounds(elb_obj):

    elb_name, elb_attached_sgs = extract_elb_name_and_sgs_from_elb(elb_obj)

    service = guess_service_name(elb_name)
    instance_sgs = get_main_security_groups(service).values()
    elb_attached_sgs.extend(instance_sgs)
    # TODO: matching all SGs with regex `name-\d` and uses them
    sgs = set(elb_attached_sgs)

    inbounds = {}
    for sg in sgs:
        inbounds.update(sgid_to_inbounds(sg))

    return inbounds


def get_services_access_elb(elb_obj):
    # TODO: use elb name instead of obj

    inbounds = elb_obj_to_all_inbounds(elb_obj)
    svcs = list(set([guess_service_name(inbound_name) for inbound_name in inbounds]))

    elb_name, _ = extract_elb_name_and_sgs_from_elb(elb_obj)
    print('{} has {} inbounds {}'.format(elb_name, len(inbounds), inbounds))

    def get_this_elb_service_name(elb_name):
        return guess_service_name(elb_name)

    this_elb_service_name = get_this_elb_service_name(elb_name)

    try:
        svcs.remove(this_elb_service_name)
    except ValueError:
        pass
    svcs.sort()
    print('{} has {} incoming connections from services {}'.format(this_elb_service_name, len(svcs), svcs))
    return svcs


def get_elb_object(elb_name):
    return [elb_obj for elb_obj in all_load_balancers if elb_obj['LoadBalancerName']]


def get_elb_info(elb_name):
    resp = get_elb_object(elb_name)

    lb_objs = resp['LoadBalancerDescriptions']
    assert len(lb_objs) == 1
    return lb_objs[0]


def group_name_contain(servicename, g):
    return servicename in g['GroupName']


def get_cluster(groupname):
    return groupname.split('-')[0]


def get_cluster_from_sg(sg):
    return get_cluster(sg['GroupName'])


def get_env(security_groups, wanted_env):

    for env, sgs in itertools.groupby(sorted(security_groups, key=lambda x: x['GroupName']), get_cluster_from_sg):
        if env == wanted_env:
            return list(sgs)
    return []


def sg_names(sgs):
    return [sg['GroupName'] for sg in sgs]


def names(list_objs):

    for key in ('GroupName', 'RoleName', 'LoadBalancerName', 'AutoScalingGroupName', 'CacheClusterId'):
        try:
            return [obj[key] for obj in list_objs]
        except KeyError:
            continue

    try:
        return [obj['Instances'][0]['InstanceId'] for obj in list_objs]
    except KeyError:
        assert False, 'Cannot get name of {}'.format(list_objs[0])


def filter_by_service_name(botoresults, choice_func):
    return list(filter(choice_func, botoresults))


def group_by(things, func):
    for prefix, g in itertools.groupby(things, func):
        yield prefix, list(g)


# TODO maybe enum type for Role, Group... Name
def first_key_prefix(item, key='RoleName'):
    return item[key].split('-')[0]


def get_asg_of_service(service_name):

    service_asgs = filter_by_service_name(
        all_auto_scaling_groups,
        lambda asg: 'prd-' in asg['AutoScalingGroupName'] and service_name == guess_service_name(asg['AutoScalingGroupName'])
    )
    assert len(service_asgs) == 1, "unhandled case {}".format(names(service_asgs))
    return service_asgs[0]


def get_asg_stats(asg):
    return {'MinSize': asg['MinSize'],
            'MaxSize': asg['MaxSize'],
            'DesiredCapacity': asg['DesiredCapacity'],
            }


def format_asg_stats(asg_stats):
    return "💻 {MinSize} <--- {DesiredCapacity} ---> {MaxSize}".format(**asg_stats)


# asg.describe_launch_configurations(LaunchConfigurationNames=[prd_asg['LaunchConfigurationName']])
getter = dict(
    asg=get_all_auto_scaling_groups,
    elasticache=get_all_cache_clusters,
    elb=get_all_load_balancers,
    sg=get_all_security_groups,
    ec2=get_all_ec2_instances,
    roles=get_all_roles
)


def main():
    # all_auto_scaling_groups = get_all_auto_scaling_groups()
    # all_load_balancers = get_all_load_balancers()
    # all_security_groups = get_all_security_groups()
    # all_ec2_instances = get_all_ec2_instances()
    # all_roles = get_all_roles()
    all_ec_clusters = get_all_cache_clusters()

    rep_groups = set()
    for i in all_ec_clusters:
        try:
            rep_groups.add(i['ReplicationGroupId'])
        except KeyError:
            pass
            # print('ERROR', i['CacheClusterId'])

    print(len(rep_groups))


if __name__ == "__main__":
    main()

==> faws-0.1.1.tar.gz: faws-0.1.1/faws/core.py <==
# -*- coding: utf-8 -*-
import itertools

import boto3

RESOURCES = ('ec2', 'elb', 'sg', 'roles', 'elasticache', 'asg')

ec2 = boto3.client('ec2')
elbc = boto3.client('elb')
iam = boto3.client('iam')
ec = boto3.client('elasticache')
asg = boto3.client('autoscaling')

all_auto_scaling_groups = []
all_ec_clusters = []
all_load_balancers = []
all_security_groups = []
all_ec2_instances = []
all_roles = []


def request_all(client, funcname, topkey, **kwargs):
    '''
    Helper function get all result, handle pagination, return list of things
    '''
    paginator = client.get_paginator(funcname)
    all_entities = []

    page_iterator = paginator.paginate(**kwargs)
    for page in page_iterator:
        all_entities.extend(page[topkey])

    return all_entities


def get_all_ec2_instances():
    all_instances = []
    for reservation in ec2.describe_instances()['Reservations']:
        all_instances.extend(reservation['Instances'])
    return all_instances


def get_all_security_groups():
    return ec2.describe_security_groups()['SecurityGroups']


def get_all_load_balancers():
    return request_all(elbc, 'describe_load_balancers', 'LoadBalancerDescriptions', PageSize=400)


def get_all_auto_scaling_groups():
    return request_all(asg, 'describe_auto_scaling_groups', 'AutoScalingGroups', MaxRecords=100)


def get_all_cache_clusters():
    return request_all(ec, 'describe_cache_clusters', 'CacheClusters')


def get_all_roles():
    # TODO paginate
    roles = iam.list_roles(MaxItems=1000)
    assert len(roles['Roles']) < 1000, 'THere are more than 1000 roles, do paginate, lazy!'
    return roles['Roles']


def get_instances():
    '''
    Get all instances, save to global var to not query AWS anymore
    '''
    pass


def get_instance(instance_id):
    '''
    Get from local instead of call AWS
    '''
    pass


def instance_ids_to_ips(instance_ids):
    # type: List -> Dict[id->ip]
    return


def get_elbs(namefilter=None):
    '''
    Get ELBs and save to global
    '''
    # Use fitler to filter base on elb name
    pass


# TODO design func to get ELB health - related instances
# HOW TO quickly get list of instances that out of service? and/or because of Unhealthy ELB check?


def guess_service_name(lb_name):
    # NOTE: order, duplication matter
    for part in ('ec2-', 'elb-', 'prd-', 'elb-', 'prod-'):
        if lb_name.startswith(part):
            lb_name = lb_name[len(part):]

    # TODO: should -int is separated thing?
    if lb_name.endswith('-int'):
        lb_name = lb_name[:-4]

    # TODO: handle suffix number, like service-1 app-2... workaround when there are too much inbound rules
    if lb_name.endswith('-1'):
        lb_name = lb_name[:-2]

    return lb_name


def extract_elb_name_and_sgs_from_elb(lb):
    if isinstance(lb, str):
        # got pretend sg name, and get inbound of the SG
        sgs = [sg for sg in all_security_groups if sg['GroupName'].lower() == lb]
        inbound_sgs = []
        for sg in sgs:
            for inbound_rule in sg['IpPermissions']:
                inbound_sg_id = inbound_rule['UserIdGroupPairs'][0]['GroupId']
                inbound_sgs.append(inbound_sg_id)

        return lb, inbound_sgs
    else:
        return lb['LoadBalancerName'], lb['SecurityGroups']


def sgid_to_inbounds(sgid):
    # inbound rules -> security group ids -> security group name
    security_group = [sg for sg in all_security_groups if sg['GroupId'] == sgid]
    assert len(security_group) == 1, 'Should only 1 SG has given ID {}'.format(sgid)

    inbound_rules = security_group[0]['IpPermissions']

    internet_access = False
    groupsids = []
    for rule in inbound_rules:
        if rule['UserIdGroupPairs']:
            groupsids.append(rule['UserIdGroupPairs'])
        else:
            # UserIdGroupPairs empty means it is a CIDR rule - maybe check for internet incoming?
            if rule['FromPort'] in (80, 443):
                # TODO maybe check more
                internet_access = True

    sgs = [g['GroupId'] for L in groupsids for g in L]
    ret = {}
    try:
        if sgs:
            ret = {sg['GroupName']: sg['GroupId']
                   for sg in all_security_groups if sg['GroupId'] in sgs}
    except Exception as e:
        print(e)

    if internet_access:
        ret.update({'INTERNET': '1'})
    return ret


def find_security_groups_contain(service_name):
    service_name = service_name.lower()
    return [sg for sg in all_security_groups if service_name in sg['GroupName'].lower()]


def is_prod_sg(sg):
    return 'prd-' in sg['GroupName'] or 'prod-' in sg['GroupName']


def get_sg_name(sg):
    return sg['GroupName']


def get_main_security_groups(service_name):
    groups = [g for g in find_security_groups_contain(service_name) if is_prod_sg(g)]

    for sg in groups:
        inbounds = sgid_to_inbounds(sg['GroupId'])
        elb = {name: inbounds[name] for name in inbounds if 'elb-' in name}
        if elb:
            elb.update(
                {sg['GroupName']: sg['GroupId']}
            )
            return elb
    return {}


def elb_obj_to_all_inbounds(elb_obj):

    elb_name, elb_attached_sgs = extract_elb_name_and_sgs_from_elb(elb_obj)

    service = guess_service_name(elb_name)
    instance_sgs = get_main_security_groups(service).values()
    elb_attached_sgs.extend(instance_sgs)
    # TODO: matching all SGs with regex `name-\d` and uses them
    sgs = set(elb_attached_sgs)

    inbounds = {}
    for sg in sgs:
        inbounds.update(sgid_to_inbounds(sg))

    return inbounds


def get_services_access_elb(elb_obj):
    # TODO: use elb name instead of obj

    inbounds = elb_obj_to_all_inbounds(elb_obj)
    svcs = list(set([guess_service_name(inbound_name) for inbound_name in inbounds]))

    elb_name, _ = extract_elb_name_and_sgs_from_elb(elb_obj)
    print('{} has {} inbounds {}'.format(elb_name, len(inbounds), inbounds))

    def get_this_elb_service_name(elb_name):
        return guess_service_name(elb_name)

    this_elb_service_name = get_this_elb_service_name(elb_name)

    try:
        svcs.remove(this_elb_service_name)
    except ValueError:
        pass
    svcs.sort()
    print('{} has {} incoming connections from services {}'.format(this_elb_service_name, len(svcs), svcs))
    return svcs


def get_elb_object(elb_name):
    return [elb_obj for elb_obj in all_load_balancers if elb_obj['LoadBalancerName']]


def get_elb_info(elb_name):
    resp = get_elb_object(elb_name)

    lb_objs = resp['LoadBalancerDescriptions']
    assert len(lb_objs) == 1
    return lb_objs[0]


def group_name_contain(servicename, g):
    return servicename in g['GroupName']


def get_cluster(groupname):
    return groupname.split('-')[0]


def get_cluster_from_sg(sg):
    return get_cluster(sg['GroupName'])


def get_env(security_groups, wanted_env):

    for env, sgs in itertools.groupby(sorted(security_groups, key=lambda x: x['GroupName']), get_cluster_from_sg):
        if env == wanted_env:
            return list(sgs)
    return []


def sg_names(sgs):
    return [sg['GroupName'] for sg in sgs]


def names(list_objs):

    for key in ('GroupName', 'RoleName', 'LoadBalancerName', 'AutoScalingGroupName', 'CacheClusterId', 'InstanceId'):
        try:
            return [obj[key] for obj in list_objs]
        except KeyError:
            continue

    raise Exception('Cannot get name of {}'.format(list_objs[0]))


def tags(tagged_object):
    return {tag['Key']: tag['Value']
            for tag in tagged_object.get('Tags', {})}


def instance_private_ip(instance):
    return instance.get('PrivateIpAddress')


def filter_by_service_name(botoresults, choice_func):
    return list(filter(choice_func, botoresults))


def group_by(things, func):
    for prefix, g in itertools.groupby(things, func):
        yield prefix, list(g)


# TODO maybe enum type for Role, Group... Name
def first_key_prefix(item, key='RoleName'):
    return item[key].split('-')[0]


def get_asg_of_service(service_name):

    service_asgs = filter_by_service_name(
        all_auto_scaling_groups,
        lambda asg: 'prd-' in asg['AutoScalingGroupName'] and service_name == guess_service_name(asg['AutoScalingGroupName'])
    )
    assert len(service_asgs) == 1, "unhandled case {}".format(names(service_asgs))
    return service_asgs[0]


def get_asg_stats(asg):
    return {'MinSize': asg['MinSize'],
            'MaxSize': asg['MaxSize'],
            'DesiredCapacity': asg['DesiredCapacity'],
            }


def format_asg_stats(asg_stats):
    return "💻 {MinSize} <--- {DesiredCapacity} ---> {MaxSize}".format(**asg_stats)


# asg.describe_launch_configurations(LaunchConfigurationNames=[prd_asg['LaunchConfigurationName']])
getter = dict(
    asg=get_all_auto_scaling_groups,
    elasticache=get_all_cache_clusters,
    elb=get_all_load_balancers,
    sg=get_all_security_groups,
    ec2=get_all_ec2_instances,
    roles=get_all_roles
)


def main():
    # all_auto_scaling_groups = get_all_auto_scaling_groups()
    # all_load_balancers = get_all_load_balancers()
    # all_security_groups = get_all_security_groups()
    # all_ec2_instances = get_all_ec2_instances()
    # all_roles = get_all_roles()
    all_ec_clusters = get_all_cache_clusters()

    rep_groups = set()
    for i in all_ec_clusters:
        try:
            rep_groups.add(i['ReplicationGroupId'])
        except KeyError:
            pass
            # print('ERROR', i['CacheClusterId'])

    print(len(rep_groups))


if __name__ == "__main__":
    main()
